@@ -55,7 +55,7 @@ my %tricks = map {; $_ => 1 }
 
 # elements that change text style
 my %elements_text_style = map {; $_ => 1 }
-  qw( body font table tr th td big small basefont marquee span p div ),
+  qw( body font table tr th td big small basefont marquee span p div a ),
 ;
 
 # elements that insert whitespace
@@ -80,13 +80,14 @@ $ok_attributes{basefont}{$_} = 1 for qw( color face size );
 $ok_attributes{body}{$_} = 1 for qw( text bgcolor link alink vlink background );
 $ok_attributes{font}{$_} = 1 for qw( color face size );
 $ok_attributes{marquee}{$_} = 1 for qw( bgcolor background );
-$ok_attributes{table}{$_} = 1 for qw( bgcolor );
-$ok_attributes{td}{$_} = 1 for qw( bgcolor );
-$ok_attributes{th}{$_} = 1 for qw( bgcolor );
-$ok_attributes{tr}{$_} = 1 for qw( bgcolor );
+$ok_attributes{table}{$_} = 1 for qw( bgcolor style );
+$ok_attributes{td}{$_} = 1 for qw( bgcolor style );
+$ok_attributes{th}{$_} = 1 for qw( bgcolor style );
+$ok_attributes{tr}{$_} = 1 for qw( bgcolor style );
 $ok_attributes{span}{$_} = 1 for qw( style );
 $ok_attributes{p}{$_} = 1 for qw( style );
 $ok_attributes{div}{$_} = 1 for qw( style );
+$ok_attributes{a}{$_} = 1 for qw( style );
 
 sub new {
   my ($class, $character_semantics_input, $character_semantics_output) = @_;
@@ -239,7 +240,7 @@ sub parse {
 
   # # (outdated claim) HTML::Parser converts &nbsp; into a question mark ("?")
   # # for some reason, so convert them to spaces.  Confirmed in 3.31, at least.
-  # ... Actually it doesn't, it is correctly coverted into Unicode NBSP,
+  # ... Actually it doesn't, it is correctly converted into Unicode NBSP,
   # nevertheless it does not hurt to treat it as a space.
   $text =~ s/&nbsp;/ /g;
 
@@ -253,11 +254,26 @@ sub parse {
       if !$self->{SA_character_semantics_input};
   } else {
     $self->SUPER::utf8_mode($self->{SA_character_semantics_input} ? 0 : 1);
+    my $utf8_mode = $self->SUPER::utf8_mode;
     dbg("message: HTML::Parser utf8_mode %s",
-        $self->SUPER::utf8_mode ? "on (assumed UTF-8 octets)"
-                                : "off (default, assumed Unicode characters)");
+        $utf8_mode ? "on (assumed UTF-8 octets)"
+                   : "off (default, assumed Unicode characters)");
   }
-  $self->SUPER::parse($text);
+
+  eval {
+    local $SIG{__WARN__} = sub {
+      my $err = $_[0];
+      $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/s;
+      info("message: HTML::Parser warning: $err");
+    };
+    $self->SUPER::parse($text);
+  };
+
+  # bug 7437: deal gracefully with HTML::Parser misbehavior on unclosed <style> and <script> tags
+  # (typically from not passing the entire message to spamc, but possibly a DoS attack)
+  $self->SUPER::parse("</style>") while exists $self->{inside}{style} && $self->{inside}{style} > 0;
+  $self->SUPER::parse("</script>") while exists $self->{inside}{script} && $self->{inside}{script} > 0;
+
   $self->SUPER::eof;
 
   return $self->{text};
@@ -493,7 +509,7 @@ sub text_style {
         $new{style} = $attr->{style};
 	my @parts = split(/;/, $new{style});
 	foreach (@parts) {
-	  if (/^\s*(background-)?color:\s*(.+)\s*$/i) {
+	  if (/^\s*(background-)?color:\s*(.+?)\s*$/i) {
 	    my $whcolor = $1 ? 'bgcolor' : 'fgcolor';
 	    my $value = lc $2;
 
@@ -504,6 +520,10 @@ sub text_style {
                                        map { !$_ ? 0 : $_ > 255 ? 255 : $_ }
                                            @rgb[0..2]);
             }
+            elsif ($value eq 'inherit') {
+              # do nothing, just prevent parsing of the valid
+              # CSS3 property value as 'invalid color' (Bug 7778)
+            }
 	    else {
 	      $new{$whcolor} = name_to_rgb($value);
 	    }
@@ -608,7 +628,16 @@ sub html_tests {
   my ($self, $tag, $attr, $num) = @_;
 
   if ($tag eq "font" && exists $attr->{face}) {
-    if ($attr->{face} !~ /^[a-z ][a-z -]*[a-z](?:,\s*[a-z][a-z -]*[a-z])*$/i) {
+    # Fixes from Bug 5956, 7312
+    # Examples seen in ham:
+    #  "Tahoma", Verdana, Arial, sans-serif
+    #  'Montserrat', sans-serif
+    #  Arial,Helvetica,Sans-Serif;
+    #  .SFUIDisplay
+    #  hirakakupro-w3
+    # TODO: There's still the problem completely foreign unicode strings,
+    # probably this rule should be deprecated.
+    if ($attr->{face} !~ /^\s*["'.]?[a-z ][a-z -]*[a-z]\d?["']?(?:,\s*["']?[a-z][a-z -]*[a-z]\d?["']?)*;?$/i) {
       $self->put_results(font_face_bad => 1);
     }
   }
@@ -742,7 +771,10 @@ sub html_text {
   my $invisible_for_bayes = 0;
 
   # NBSP:  UTF-8: C2 A0, ISO-8859-*: A0
-  if ($text !~ /^(?:[ \t\n\r\f\x0b]|\xc2\xa0)*\z/s) {
+  # Bug 7374 - regex recursion limit exceeded
+  #if ($text !~ /^(?:[ \t\n\r\f\x0b]|\xc2\xa0)*\z/s) {
+  # .. alternative way, remove from string and see if there's anything left
+  if (do {(my $tmp = $text) =~ s/(?:[ \t\n\r\f\x0b]|\xc2\xa0)//gs; length($tmp)}) {
     $invisible_for_bayes = $self->html_font_invisible($text);
   }
 
